/* This file is part of brainsuck.
 *
 * brainsuck is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * brainsuck is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with brainsuck; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>

#define DBG(str)	
/*
#define DBG(str)			printf("BF#%d,",ftell(fsrc));puts(str);putchar('\n');
*/

#define BFMEM_LEN 				4096U

#define PRGAUTHOR				"Poke&Logout on Shucon2007"
#define PRGNAME					"brainsuck"
#define PRGVER					"1.01"

#define ERR_OK					0
#define ERR_BAD_NUM 			-1
#define ERR_SRC_FILE 			-2
#define ERR_STACK_OVERFL		-3
#define ERR_UNEX_EOF			-4
#define ERR_GETPOS_SRC		-5
#define ERR_SETPOS_SRC		-6
#define ERR_STACK_UNDERFL	-7
#define ERR_MEMDUMP_FILE		-8

#define BF_INC_PTR				'>'
#define BF_DEC_PTR				'<'
#define BF_INC_BYTE			'+'
#define BF_DEC_BYTE			'-'
#define BF_PUT_BYTE			'.'
#define BF_GET_BYTE			','
#define BF_LOOP_BEG			'['
#define BF_LOOP_END			']'

#define STACK_LOOP_LEN		64

static unsigned char mem[BFMEM_LEN];
static unsigned int ptr = 0;
static fpos_t loopstack[STACK_LOOP_LEN];
static int stacktop = 0;

void DisplayLicense(void)
	{
	printf(
PRGNAME" (brainfuck interpreter) ver."PRGVER", build ("__DATE__","__TIME__").\n\n"
"   Author: "PRGAUTHOR".\n"
"   This program is free software; you can redistribute it and/or modify\n"
"   it under the terms of the GNU General Public License as published by\n"
"   the Free Software Foundation; either version 2, or (at your option)\n"
"   any later version.\n"
"\n"
"   This program is distributed in the hope that it will be useful,\n"
"   but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
"   GNU General Public License for more details.\n"
"\n"
"   You should have received a copy of the GNU General Public License\n"
"   along with this program; if not, write to the Free Software\n"
"   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\n"
	);
	}

void DumpMemory(FILE *stream, unsigned char *mem)
	{
	unsigned int ptr = 0;
	int n;
	while(ptr < BFMEM_LEN)
		{
		fprintf(stream, "%04x:",ptr);
		for(n = 0; (n < 16) && (ptr < BFMEM_LEN); n++, ptr++)
			fprintf(stream,  " %02x", mem[ptr]);
		fputc('\n',stream);
		}
	}

int main(int argn, char* argv[])
	{
	FILE *fsrc = NULL;
	int cmd;

	if( (argn != 2) && (argn != 3) )
		{
		DisplayLicense();
		fprintf(stderr, PRGNAME": error: Bad number of parameters.\n");
		printf("\nUsage: \n\t%s <bfsourcefile.bfc> [memorydump.txt]\n",argv[0]);
		return(ERR_BAD_NUM);
		}

	if((fsrc = fopen(argv[1],"rt")) == NULL)
		{
		fprintf(stderr, PRGNAME": error: Cannot open source file.\n");
		return(ERR_SRC_FILE);
		}

	while( (cmd = fgetc(fsrc)) != EOF)
	{
	switch(cmd)
		{
		case(BF_INC_PTR):
			DBG("BF_INC_PTR");
			ptr++;
			ptr %= BFMEM_LEN;
		break; /* BF_INC_PTR */

		case(BF_DEC_PTR):
			DBG("BF_DEC_PTR");
			ptr--;
			ptr %= BFMEM_LEN;
		break; /* BF_DEC_PTR */

		case(BF_INC_BYTE):
			DBG("BF_INC_BYTE");
			mem[ptr]++;
		break; /* BF_INC_BYTE */

		case(BF_DEC_BYTE):
			DBG("BF_DEC_BYTE");
			mem[ptr]--;
		break; /* BF_DEC_BYTE */

		case(BF_PUT_BYTE):
			DBG("BF_PUT_BYTE");
			putchar(mem[ptr]);
/*			printf("ptr = %d *ptr=%d\n", ptr, mem[ptr]); */
		break; /* BF_PUT_BYTE */

		case(BF_GET_BYTE):
			DBG("BF_GET_BYTE");
			mem[ptr] = (unsigned char)getchar();
		break; /* BF_GET_BYTE */

		case(BF_LOOP_BEG):
			DBG("BF_LOOP_BEG");
			if(mem[ptr])
				{ /* push to the stack */
				if(stacktop > STACK_LOOP_LEN)
					{
					fprintf(stderr, PRGNAME": error: Loop stack overflow, too much inner loops.\n");
					fclose(fsrc);
					return(ERR_STACK_OVERFL);
					}
				if( fgetpos(fsrc, &(loopstack[stacktop])) )
					{
					fprintf(stderr, PRGNAME": error: Cannot get position in the source file.\n");
					fclose(fsrc);
					return(ERR_GETPOS_SRC);
					}
				stacktop++;
				}
			else
				{ /* ignore loop seek for BF_LOOP_END */
				do
					{
					if(feof(fsrc))
						{
						fprintf(stderr, PRGNAME": error: Unexpected end of file, missing end of the loop.\n");
						fclose(fsrc);
						return(ERR_UNEX_EOF);
						}
					}
				while(fgetc(fsrc) != BF_LOOP_END);
				}
		break; /* BF_LOOP_BEG */

		case(BF_LOOP_END):
			DBG("BF_LOOP_END");
/*			printf("ptr = %d *ptr=%d\n", ptr, mem[ptr]); */
			if(stacktop < 1)
				{
				fprintf(stderr, PRGNAME": error: Loop stack underflow, missing begin of the loop.\n");
				fclose(fsrc);
				return(ERR_STACK_UNDERFL);
				}

			if(mem[ptr] != 0)
				{ /* pop from the stack */
				DBG("BF_LOOP_COND_YES");
				if( fsetpos(fsrc, &(loopstack[stacktop-1])) )
					{
					fprintf(stderr, PRGNAME": error: Cannot set position in the source file.\n");
					fclose(fsrc);
					return(ERR_SETPOS_SRC);
					}
				}
			else
				stacktop--;

		break; /* BF_LOOP_END */

		default:
			DBG("COMMENT");
		break;
		} /* switch */
	} /* while */

	fclose(fsrc);
	fflush(stdout);

	if(argn == 3)
		{
		if((fsrc = fopen(argv[2],"wt")) == NULL)
			{
			fprintf(stderr, PRGNAME": error: Cannot create memory dump file.\n");
			return(ERR_MEMDUMP_FILE);
			}
		DumpMemory(fsrc, mem);
		fclose(fsrc);
		}

	return(ERR_OK);
	} /* main() */
